/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<errno.h>
#include	<fcntl.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/dirent.h>
#include	"lxrun.h"
#include	"linux/fcntl.h"
#include	"devmap.h"

/*
 * file i/o system calls
 */

/*
 *	open
 *	fcntl
 *	flock		(NOT IMPLEMENTED)
 *	__llseek
 *	readdir
 *	fdatasync	(partially implemented)
 */

struct
{
	int	lx_flag;
	int	os_flag;
} oflags[] =
{
	{ LX_O_CREAT,	O_CREAT		},
	{ LX_O_EXCL,	O_EXCL		},
	{ LX_O_NOCTTY,	O_NOCTTY	},
	{ LX_O_TRUNC,	O_TRUNC		},
	{ LX_O_APPEND,	O_APPEND	},
	{ LX_O_NONBLOCK,O_NONBLOCK	},
	{ LX_O_SYNC,	O_SYNC		},
};

static int
copy_in_oflags(int *from, int *to)
{
	int	i;
	int	lx_flags;
	int	os_flags;

	lx_flags = *from;
	switch (lx_flags & LX_O_ACCMOD)
	{
		case LX_O_RDONLY:
			os_flags = O_RDONLY;
			break;
		case LX_O_WRONLY:
			os_flags = O_WRONLY;
			break;
		case LX_O_RDWR:
			os_flags = O_RDWR;
			break;
		default:
			errno = EINVAL;
			return -1;
	}

	lx_flags &= ~LX_O_ACCMOD;

	for (i = 0; i < sizeof(oflags) / sizeof(oflags[0]); i++)
	{
		if (lx_flags == 0)
			break;

		if (oflags[i].lx_flag & lx_flags)
		{
			lx_flags &= ~oflags[i].lx_flag;
			os_flags |=  oflags[i].os_flag;
		}
	}

	if (lx_flags != 0)
	{
		errno = EINVAL;
		return -1;
	}

	*to = os_flags;

	return 0;
}

static int
copy_out_oflags(int *from, int *to)
{
	int	i;
	int	lx_flags;
	int	os_flags;

	os_flags = *from;
	switch (os_flags & O_ACCMODE)
	{
		case O_RDONLY:
			lx_flags = LX_O_RDONLY;
			break;
		case O_WRONLY:
			lx_flags = LX_O_WRONLY;
			break;
		case O_RDWR:
			lx_flags = LX_O_RDWR;
			break;
		default:
			errno = EINVAL;
			return -1;
	}

	os_flags &= ~O_ACCMODE;

	for (i = 0; i < sizeof(oflags) / sizeof(oflags[0]); i++)
	{
		if (os_flags == 0)
			break;

		if (oflags[i].os_flag & os_flags)
		{
			os_flags &= ~oflags[i].os_flag;
			lx_flags |=  oflags[i].lx_flag;
		}
	}

	if (os_flags != 0)
	{
		errno = EINVAL;
		return -1;
	}

	*to = lx_flags;

	return 0;
}

static int
copy_in_flock(struct lx_flock *from, struct flock *to)
{
	if (! verify_read(from, sizeof(*from)))
		return -1;

	switch (from->l_type)
	{
		case LX_F_RDLCK:
			to->l_type = F_RDLCK;
			break;
		case LX_F_WRLCK:
			to->l_type = F_WRLCK;
			break;
		case LX_F_UNLCK:
			to->l_type = F_UNLCK;
			break;
		default:
			errno = EINVAL;
			return -1;
	}

	to->l_whence	= from->l_whence;
	to->l_start	= from->l_start;
	to->l_len	= from->l_len;
	to->l_pid	= from->l_pid;

	return 0;
}

static int
copy_out_flock(struct flock *from, struct lx_flock *to)
{
	if (! verify_write(to, sizeof(*to)))
		return -1;

	switch (from->l_type)
	{
		case F_RDLCK:
			to->l_type = LX_F_RDLCK;
			break;
		case F_WRLCK:
			to->l_type = LX_F_WRLCK;
			break;
		case F_UNLCK:
			to->l_type = LX_F_UNLCK;
			break;
		default:
			errno = EINVAL;
			return -1;
	}

	to->l_whence	= from->l_whence;
	to->l_start	= from->l_start;
	to->l_len	= from->l_len;
	to->l_pid	= from->l_pid;

	return 0;
}

int
lx_open(char *lx_path, int lx_flags, int mode)
{
	char	os_path[PATH_MAX];
	int	flags;

	if (lx_pathname(lx_path, os_path, PATH_MAX) < 0)
		return -1;

	if (copy_in_oflags(&lx_flags, &flags) != 0)
		return -1;

	return open(os_path, flags, mode);
}

int
lx_close(int fd)
{
	/* invalidate this file descriptor in the devmap cache */
	devmap_invalidate(fd);

	return close(fd);
}

int
lx_read(int fd, void *buf, int len)
{
	return read(fd, buf, len);
}

int
lx_write(int fd, void *buf, int len)
{
	int	n;

	n = write(fd, buf, len);

	if (errno == EWOULDBLOCK && len > 4096)	/* kludge for TCP/IP */
	{
		errno = 0;
		n = write(fd, buf, 4096);
	}

	return n;
}

int
lx_fcntl(int fd, int lx_cmd, int arg)
{
	int		r;
	int		flags;
	struct	flock	f;

	switch (lx_cmd)
	{
		case LX_F_DUPFD:
			return fcntl(fd, F_DUPFD, arg);

		case LX_F_GETFD:
			return fcntl(fd, F_GETFD, arg);

		case LX_F_SETFD:
			return fcntl(fd, F_SETFD, arg);

		case LX_F_GETFL:
			flags = -1;
			if ((r = fcntl(fd, F_GETFL, arg)) != -1)
				(void)copy_out_oflags(&r, &flags);
			return flags;

		case LX_F_SETFL:
			if (copy_in_oflags(&arg, &flags) != 0)
				return -1;
			return fcntl(fd, F_SETFL, flags);

		case LX_F_GETLK:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			if ((r = fcntl(fd, F_GETLK, &f)) == -1)
				return -1;
			if (copy_out_flock(&f, (struct lx_flock *)arg) != 0)
				return -1;
			return r;

		case LX_F_SETLK:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			return fcntl(fd, F_SETLK, &f);

		case LX_F_SETLKW:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			return fcntl(fd, F_SETLKW, &f);

		case LX_F_GETOWN:
		case LX_F_SETOWN:
			break;
	}
	errno = EINVAL;
	return -1;
}

int lx_flock()		{ errno = ENOSYS; return -1; }	

int
lx__llseek(int fd, long off_hi, long off_lo, long *res, int whence)
{
	long	l;

	if ((off_lo >= 0 && off_hi != 0) || (off_lo < 0 && off_hi != -1))
		errno = EINVAL;
	else
		l = lseek(fd, off_lo, whence);

	if (errno != 0 || ! verify_write(res, 2 * sizeof(res[0])))
		return -1;

	res[0] = l;
	res[1] = (l < 0) ? -1 : 0;

	return 0;
}	

struct lx_dirent
{
	long	d_ino;
	long	d_off;
	short	d_namlen;
	char	d_name[1];
};

int
lx_readdir(int fd, struct lx_dirent *lx_dir, int count)
{
	char		buf[512];
	struct dirent	*d	= (struct dirent *)buf;

	if (getdents(fd, d, 512) <= 0)
		return -1;

	lx_dir->d_ino	= d->d_ino;
	lx_dir->d_off	= d->d_off;
	lx_dir->d_namlen= strlen(d->d_name);
	strcpy(lx_dir->d_name, d->d_name);
	lseek(fd, d->d_off, 0);

	return d->d_reclen;
}	


/*
** The behavior of fdatasync and fsync are slightly different, so
** this isn't a perfect implementation.
** The latter changes the access time in the inode.
** The former updates the file _without_ touching the inode.
** The main advantage of fdatasync over fsync is performance.
** Hopefully, this is the _only_ difference as far as most Linux
** programs are concerned.
*/
int
lx_fdatasync(int fd)
{
	return fsync(fd);
}
