/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<assert.h>
#include	<errno.h>
#include	<signal.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/wait.h>
#include	<sys/siginfo.h>
#include	<sys/regset.h>
#include	<sys/reg.h>
#include	<ucontext.h>
#include	"lxrun.h"
#include	"linux/resource.h"
#include	"linux/signal.h"
#include	"linux/sigcontext.h"
#include	"linux/wait.h"

#if defined(UNIXWARE)
#define	sa_sigaction	sa_handler
#endif

/*
 * signals[] is a table of corresponding Linux and System V
 * error numbers - it is used by init_signals() to build the
 * lx_signo[] and os_signo[] lookup tables
 */
#define	SIG(name)	{ SIG##name, LX_SIG##name }

struct
{
	int	os_sig;
	int	lx_sig;
} signals[NSIG] =
{
	SIG(HUP),	SIG(INT),	SIG(QUIT),	SIG(ILL),
	SIG(TRAP),	SIG(ABRT),	SIG(BUS),	SIG(FPE),
	SIG(KILL),	SIG(USR1),	SIG(SEGV),	SIG(USR2),
	SIG(PIPE),	SIG(ALRM),	SIG(TERM),	SIG(CHLD),
	SIG(CONT),	SIG(STOP),	SIG(TSTP),	SIG(TTIN),
	SIG(TTOU),	SIG(XCPU),	SIG(XFSZ),	SIG(VTALRM),
	SIG(PROF),	SIG(WINCH),	SIG(PWR),
#if defined(UNIXWARE)
							SIG(URG),
	SIG(IO),	0, 		0,		0
#else
							0,
	0,		0,		0,		0
#endif
};

static int		lx_signo[NSIG+1];
static int		os_signo[NSIG+1];
static struct sigaction os_actions[NSIG+1];

int
init_signals()
{
	int			i;
	struct sigaction	s;

	extern void syscall_handler(int sig, siginfo_t *info, void *context);

	/*
	 * initialise the signal number lookup tables
	 */
	for (i = 0; i < NSIG; i++)
	{
		int	os_sig	= signals[i].os_sig;
		int	lx_sig	= signals[i].lx_sig;

		assert(os_sig >= 0 && os_sig <= NSIG);
		assert(lx_sig >= 0 && lx_sig <= LX_NSIG);

		lx_signo[os_sig]	= lx_sig;
		os_signo[lx_sig]	= os_sig;
	}

	/*
	 * install the system call handler for the int 0x80 instruction
	 * (which causes a general protection trap and results in SIGSEGV)
	 */
	sigemptyset(&s.sa_mask);
	s.sa_sigaction	= syscall_handler;
	s.sa_flags	= SA_NODEFER | SA_SIGINFO;

	return sigaction(SIGSEGV, &s, &os_actions[SIGSEGV]);
}


static int
copy_in_sigmask(lx_sigset_t *from, sigset_t *to)
{
	int		i;
	unsigned long	lx_mask;

	if (! verify_read(from, sizeof(*from)))
		return -1;

	lx_mask = *from;
	sigemptyset(to);

	for (i = 1; lx_mask != 0; i++, lx_mask >>= 1)
		if ((lx_mask & 1)  && os_signo[i] != 0)
			sigaddset(to, os_signo[i]);

	return 0;
}


static int
copy_out_sigmask(sigset_t *from, lx_sigset_t *to)
{
	int		i;
	unsigned long	lx_mask;

	if (! verify_write(to, sizeof(*to)))
		return -1;

	for (i = 0; i < NSIG && signals[i].os_sig != 0; i++)
		if (sigismember(from, signals[i].os_sig))
			lx_mask |=  1 << (signals[i].lx_sig-1);

	*to = lx_mask;

	return 0;
}

void
deliver_signal(
	int		sig,
	siginfo_t	*info,
	void		*context
	)
{
	int			save		= errno;
	struct sigaction	*act		= &os_actions[sig];
	unsigned long		*regs;
	struct lx_sigcontext_struct	lx_context;
	extern int		sendsig(void (*)(), int, void *);

#if defined(OSR5)
	regs	= (unsigned long *)((ucontext_t *)context)->uc_mcontext.regs;
#else
	regs	= (unsigned long *)((ucontext_t *)context)->uc_mcontext.gregs;
#endif

#if defined(TRACE)
	trace_signal(sig, info, context);
#endif
	if (act->sa_handler == SIG_DFL || act->sa_handler == SIG_IGN)
	{
		sigaction(sig, act, NULL);
	}
	else
	{
		/*
		 * set up Linux signal handler stack frame
		 */
		lx_context.gs		= regs[GS];
		lx_context.fs		= regs[FS];
		lx_context.es		= regs[ES];
		lx_context.ds		= regs[DS];
		lx_context.edi		= regs[EDI];
		lx_context.esi		= regs[ESI];
		lx_context.ebp		= regs[EBP];
		lx_context.esp		= regs[ESP];
		lx_context.ebx		= regs[EBX];
		lx_context.edx		= regs[EDX];
		lx_context.ecx		= regs[ECX];
		lx_context.eax		= regs[EAX];
		lx_context.trapno	= 0;
		lx_context.err		= 0;
		lx_context.eip		= regs[EIP];
		lx_context.cs		= regs[CS];
		lx_context.eflags	= regs[EFL];
		lx_context.esp_at_signal= regs[ESP];
		lx_context.ss		= regs[SS];
		lx_context.i387		= 0;
		lx_context.oldmask	= 0;
		lx_context.cr2	= 0;

		sendsig(act->sa_handler, lx_signo[sig], &lx_context);

		/*
		 * copy back any changes
		 */
#if defined(TRACE)

#define	CHECK(REG, reg)	\
		if (regs[REG] != lx_context.reg) \
			tprintf(#REG " changed from %8x to %8x\n", \
				regs[REG], lx_context.reg)

		CHECK(EDI, edi);
		CHECK(ESI, esi);
		CHECK(EBP, ebp);
		CHECK(ESP, esp);
		CHECK(EBX, ebx);
		CHECK(EDX, edx);
		CHECK(ECX, ecx);
		CHECK(EAX, eax);
		CHECK(EIP, eip);
		CHECK(EFL, eflags);
#endif

		regs[EDI]	= lx_context.edi;
		regs[ESI]	= lx_context.esi;
		regs[EBP]	= lx_context.ebp;
		regs[ESP]	= lx_context.esp;
		regs[EBX]	= lx_context.ebx;
		regs[EDX]	= lx_context.edx;
		regs[ECX]	= lx_context.ecx;
		regs[EAX]	= lx_context.eax;
		regs[EIP]	= lx_context.eip;
		regs[EFL]	= lx_context.eflags;
	}

	errno = save;
}


int
lx_sigaction(int lx_sig, struct lx_sigaction *lx_new_act, struct lx_sigaction *lx_old_act)
{
	int			sig;
	struct sigaction	new_action;
	struct sigaction	old_action;
	struct sigaction	*new_act	= NULL;
	struct sigaction	*old_act	= NULL;

	if (lx_sig <= 0 || lx_sig > LX_NSIG)
	{
		errno = EINVAL;
		return -1;
	}

	sig = os_signo[lx_sig];

	if (lx_new_act)
	{
		if (! verify_read(lx_new_act, sizeof(*lx_new_act)))
			return -1;

		new_act	= &new_action;

		if (lx_new_act->handler == LX_SIG_DFL)
			new_act->sa_handler = SIG_DFL;
		else if (lx_new_act->handler == LX_SIG_IGN)
			new_act->sa_handler = SIG_IGN;
		else
			new_act->sa_sigaction = deliver_signal;

		copy_in_sigmask(&lx_new_act->mask, &new_act->sa_mask);
		/*
		 * make sure SIGSEGV is never blocked
		 */
		sigdelset(&new_act->sa_mask, SIGSEGV);

		new_act->sa_flags	= SA_SIGINFO;
		if (lx_new_act->flags & LX_SA_NOCLDSTOP)
			new_act->sa_flags	|= SA_NOCLDSTOP;
		if (lx_new_act->flags & LX_SA_STACK)
			new_act->sa_flags	|= SA_ONSTACK;
		if (lx_new_act->flags & LX_SA_RESTART)
			new_act->sa_flags	|= SA_RESTART;
		if (lx_new_act->flags & LX_SA_ONESHOT)
			new_act->sa_flags	|= SA_NODEFER;
	}

	if (lx_old_act)
		old_act	= &old_action;

	/*
	 * if signal is SIGSEGV just retrieve the old action
	 * from the emulator's copy, but don't actually call
	 * sigaction()
	 */
	if (sig == SIGSEGV)
	{
		if (old_act)
			*old_act = os_actions[sig];
	}
	else
	{
		if (sigaction(sig, new_act, old_act) != 0)
			return -1;
	}

	if (old_act)
	{
		if (! verify_write(lx_old_act, sizeof(*lx_old_act)))
			return -1;

		if (old_act->sa_handler == SIG_DFL)
			lx_old_act->handler = LX_SIG_DFL;
		else if (old_act->sa_handler = SIG_IGN)
			lx_old_act->handler = LX_SIG_IGN;
		else
			lx_old_act->handler = os_actions[sig].sa_handler;

		copy_out_sigmask(&old_act->sa_mask, &lx_old_act->mask);

		lx_old_act->flags	= 0;
		if (old_act->sa_flags & SA_NOCLDSTOP)
			lx_old_act->flags	|= LX_SA_NOCLDSTOP;
		if (old_act->sa_flags & SA_ONSTACK)
			lx_old_act->flags	|= LX_SA_STACK;
		if (old_act->sa_flags & SA_RESTART)
			lx_old_act->flags	|= LX_SA_RESTART;
		if (old_act->sa_flags & SA_NODEFER)
			lx_old_act->flags	|= LX_SA_ONESHOT;
	}

	/*
	 * finally update the current action for this signal
	 */
	if (new_act)
	{
		if (new_act->sa_sigaction == deliver_signal)
			new_act->sa_handler = lx_new_act->handler;

		os_actions[sig] = *new_act;
	}

	return 0;
}	


int
lx_signal(int lx_sig, lx_sighandler_t lx_handler)
{
	struct lx_sigaction	new, old;

	new.handler	= lx_handler;
	new.mask	= 0;
	new.flags	= 0;

	if (lx_sigaction(lx_sig, &new, &old) == 0)
		return (int) old.handler;

	return -1;
}	


int
lx_sigreturn()
{
	errno = ENOSYS;
	return -1;
}	


/*
 * Linux sigsuspend() is weird ....
 */
int
lx_sigsuspend(int restart, unsigned long lx_oldmask, unsigned long lx_set)
{
	sigset_t	mask;

	copy_in_sigmask(&lx_set, &mask);
	/*
	 * make sure SIGSEGV is never blocked
	 */
	sigdelset(&mask, SIGSEGV);

	return sigsuspend(&mask);
}	


int
lx_sigpending(lx_sigset_t *lx_mask)
{
	sigset_t	mask;
	int		r;

	if ((r = sigpending(&mask)) == 0)
		copy_out_sigmask(&mask, lx_mask);

	return r;

}	


int
lx_sigprocmask(int lx_cmd, lx_sigset_t *lx_new_mask, lx_sigset_t *lx_old_mask)
{
	int		cmd		= 0;
	sigset_t	omask;
	sigset_t	nmask;
	sigset_t	*old_mask	= NULL;
	sigset_t	*new_mask	= NULL;


	if (lx_new_mask)
	{
		switch (lx_cmd)
		{
			case LX_SIG_BLOCK:
				cmd = SIG_BLOCK;
				break;
			case LX_SIG_UNBLOCK:
				cmd = SIG_UNBLOCK;
				break;
			case LX_SIG_SETMASK:
				cmd = SIG_SETMASK;
				break;
			default:
				errno = EINVAL;
				return -1;
		}

		new_mask = &nmask;

		if (copy_in_sigmask(lx_new_mask, new_mask) != 0)
			return -1;
		/*
		 * make sure SIGSEGV is never blocked
		 */
		sigdelset(new_mask, SIGSEGV);
	}

	if (lx_old_mask)
		old_mask = &omask;

	if (sigprocmask(cmd, new_mask, old_mask) != 0)
		return -1;

	if (lx_old_mask && copy_out_sigmask(old_mask, lx_old_mask) != 0)
		return -1;

	return 0;
}


int
lx_sgetmask()
{
	lx_sigset_t	omask;

	lx_sigprocmask(LX_SIG_SETMASK, NULL, &omask);

	return omask;
}


int
lx_ssetmask(lx_sigset_t nmask)
{
	lx_sigset_t	omask;

	lx_sigprocmask(LX_SIG_SETMASK, &nmask, &omask);

	return omask;
}


int
lx_kill(int pid, int lx_sig)
{
	return kill(pid, os_signo[lx_sig]);
}


int
lx_wait4(int pid, int *lx_stat_loc, int lx_options, struct lx_rusage *lx_ru)
{
	int		options	= 0;
	int		r;
	int		status;

	if (lx_options & LX_WNOHANG)
		options |= WNOHANG;
	if (lx_options & LX_WUNTRACED)
		options |= WUNTRACED;

	r = waitpid(pid, &status, options);

	/*
	 * convert the status to Linux format
	 * the only real difference here is the signal #
	 */
	if (r != -1 && lx_stat_loc)
	{
		int	lx_status = 0;

		if (WIFEXITED(status))
			lx_status = WEXITSTATUS(status) << 8;
		else if (WIFSIGNALED(status))
		{
			lx_status = lx_signo[WTERMSIG(status)];
			if (WCOREDUMP(status))
				lx_status |= 0x80;
		}
		else if (WIFSTOPPED(status))
			lx_status = (lx_signo[WSTOPSIG(status)] << 8) | 0x7f;

		if (! verify_write(lx_stat_loc, sizeof(*lx_stat_loc)))
			return -1;

		*lx_stat_loc = lx_status;
	}

	if (lx_ru)	/* resource usage NOT IMPLEMENTED */
	{
		if (! verify_write(lx_ru, sizeof(*lx_ru)))
			return -1;

		memset(lx_ru, 0, sizeof(*lx_ru));
	}

	return r;
}	


int
lx_waitpid(int pid, int *lx_stat_loc, int lx_options)
{ 
	return lx_wait4(pid, lx_stat_loc, lx_options, NULL);
}
